# File generated by contrib/scrape-ec2-sizes.py script - DO NOT EDIT manually
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

INSTANCE_TYPES = {
    "a1.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "a1.2xlarge",
            "instancesku": "D5XG23APR65M5SYE",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "a1.2xlarge",
        "name": "a1.2xlarge",
        "ram": 16384
    },
    "a1.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "a1.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "a1.4xlarge",
        "name": "a1.4xlarge",
        "ram": 32768
    },
    "a1.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "a1.large",
            "instancesku": "GC2JN6QM5JNRYDST",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "a1.large",
        "name": "a1.large",
        "ram": 4096
    },
    "a1.medium": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "a1.medium",
            "instancesku": "899RCWM3249A3E8S",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "a1.medium",
        "name": "a1.medium",
        "ram": 2048
    },
    "a1.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "a1.xlarge",
            "instancesku": "44BG498Y4VQ5GM28",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "a1.xlarge",
        "name": "a1.xlarge",
        "ram": 8192
    },
    "c1.medium": {
        "bandwidth": None,
        "disk": 350,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "5",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Compute optimized",
            "instanceType": "c1.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1.7 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 350 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c1.medium",
        "name": "c1.medium",
        "ram": 1740
    },
    "c1.xlarge": {
        "bandwidth": None,
        "disk": 1680,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "20",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Compute optimized",
            "instanceType": "c1.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "7 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 420 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c1.xlarge",
        "name": "c1.xlarge",
        "ram": 7168
    },
    "c3.2xlarge": {
        "bandwidth": None,
        "disk": 160,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "No",
            "ecu": "28",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c3.2xlarge",
            "instancesku": "YU65576U57RYW2SF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2680 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 80 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c3.2xlarge",
        "name": "c3.2xlarge",
        "ram": 15360
    },
    "c3.4xlarge": {
        "bandwidth": None,
        "disk": 320,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "No",
            "ecu": "55",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c3.4xlarge",
            "instancesku": "9ZJ8WHFGWEYGH39N",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2680 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 160 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "c3.4xlarge",
        "name": "c3.4xlarge",
        "ram": 30720
    },
    "c3.8xlarge": {
        "bandwidth": 10,
        "disk": 640,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "No",
            "ecu": "108",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c3.8xlarge",
            "instancesku": "FKFNCVGF8F4VBXQ4",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "60 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2680 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 320 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "c3.8xlarge",
        "name": "c3.8xlarge",
        "ram": 61440
    },
    "c3.large": {
        "bandwidth": None,
        "disk": 32,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "No",
            "ecu": "7",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c3.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "3.75 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2680 v2 (Ivy Bridge)",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 16 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c3.large",
        "name": "c3.large",
        "ram": 3840
    },
    "c3.xlarge": {
        "bandwidth": None,
        "disk": 80,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "No",
            "ecu": "14",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c3.xlarge",
            "instancesku": "X2KK4E7Y22HQUHTD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "7.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2680 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 40 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "c3.xlarge",
        "name": "c3.xlarge",
        "ram": 7680
    },
    "c4.2xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1000 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c4.2xlarge",
            "instancesku": "3CYNCTZXWFFZCJ4T",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2666 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c4.2xlarge",
        "name": "c4.2xlarge",
        "ram": 15360
    },
    "c4.4xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2000 Mbps",
            "ecu": "62",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c4.4xlarge",
            "instancesku": "PYDH5P62F8KBF85B",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2666 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "c4.4xlarge",
        "name": "c4.4xlarge",
        "ram": 30720
    },
    "c4.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4000 Mbps",
            "ecu": "132",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c4.8xlarge",
            "instancesku": "4VRUFQGZS9QFSB7E",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "60 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2666 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "36",
            "vpcnetworkingsupport": "true"
        },
        "id": "c4.8xlarge",
        "name": "c4.8xlarge",
        "ram": 61440
    },
    "c4.large": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "500 Mbps",
            "ecu": "8",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c4.large",
            "instancesku": "3N2AT4Z2NBKAT2N4",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "3.75 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2666 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c4.large",
        "name": "c4.large",
        "ram": 3840
    },
    "c4.xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "750 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c4.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "7.5 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2666 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "c4.xlarge",
        "name": "c4.xlarge",
        "ram": 7680
    },
    "c5.12xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "188",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.12xlarge",
            "instancesku": "B2D6JZJWM7HPMFFP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Deep Learning Boost; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5.12xlarge",
        "name": "c5.12xlarge",
        "ram": 98304
    },
    "c5.18xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "281",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.18xlarge",
            "instancesku": "H2XX5A4Y823VVMXZ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "144 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "144",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "72",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5.18xlarge",
        "name": "c5.18xlarge",
        "ram": 147456
    },
    "c5.24xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "375",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.24xlarge",
            "instancesku": "GB2CVQJXDU9FR8GT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Deep Learning Boost; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5.24xlarge",
        "name": "c5.24xlarge",
        "ram": 196608
    },
    "c5.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "39",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5.2xlarge",
        "name": "c5.2xlarge",
        "ram": 16384
    },
    "c5.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "73",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5.4xlarge",
        "name": "c5.4xlarge",
        "ram": 32768
    },
    "c5.9xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "139",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.9xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "72 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "72",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "36",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5.9xlarge",
        "name": "c5.9xlarge",
        "ram": 73728
    },
    "c5.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.large",
            "instancesku": "DP5WHNEA4VF9ZKK3",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5.large",
        "name": "c5.large",
        "ram": 4096
    },
    "c5.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "20",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.xlarge",
            "instancesku": "2QD5QA8MDXRPPAKG",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5.xlarge",
        "name": "c5.xlarge",
        "ram": 8192
    },
    "c5a.12xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5a.12xlarge",
        "name": "c5a.12xlarge",
        "ram": 98304
    },
    "c5a.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6300 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.16xlarge",
            "instancesku": "U4MEJNF3728KAVBF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5a.16xlarge",
        "name": "c5a.16xlarge",
        "ram": 131072
    },
    "c5a.24xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.24xlarge",
            "instancesku": "DPHJ9PS8GF7ZRFEF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5a.24xlarge",
        "name": "c5a.24xlarge",
        "ram": 196608
    },
    "c5a.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.2xlarge",
            "instancesku": "78USCN84VB25CP2Y",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5a.2xlarge",
        "name": "c5a.2xlarge",
        "ram": 16384
    },
    "c5a.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.4xlarge",
            "instancesku": "FXZCY68C879B2BXD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5a.4xlarge",
        "name": "c5a.4xlarge",
        "ram": 32768
    },
    "c5a.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.8xlarge",
            "instancesku": "5PMN73A4GCYVKWT5",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5a.8xlarge",
        "name": "c5a.8xlarge",
        "ram": 65536
    },
    "c5a.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5a.large",
        "name": "c5a.large",
        "ram": 4096
    },
    "c5a.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5a.xlarge",
        "name": "c5a.xlarge",
        "ram": 8192
    },
    "c5ad.12xlarge": {
        "bandwidth": 12,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.12xlarge",
            "instancesku": "AMC8CQE4QGF9WSKJ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5ad.12xlarge",
        "name": "c5ad.12xlarge",
        "ram": 98304
    },
    "c5ad.16xlarge": {
        "bandwidth": 20,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6300 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1200 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5ad.16xlarge",
        "name": "c5ad.16xlarge",
        "ram": 131072
    },
    "c5ad.24xlarge": {
        "bandwidth": 20,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.24xlarge",
            "instancesku": "W45YTYFWCBGWJGSQ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5ad.24xlarge",
        "name": "c5ad.24xlarge",
        "ram": 196608
    },
    "c5ad.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.2xlarge",
            "instancesku": "D8MQ8PRJMKXJHGCR",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5ad.2xlarge",
        "name": "c5ad.2xlarge",
        "ram": 16384
    },
    "c5ad.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.4xlarge",
            "instancesku": "TQMPFTV5HMXTXSZJ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5ad.4xlarge",
        "name": "c5ad.4xlarge",
        "ram": 32768
    },
    "c5ad.8xlarge": {
        "bandwidth": 10,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.8xlarge",
            "instancesku": "X2DT43NU2ZYNA6R3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5ad.8xlarge",
        "name": "c5ad.8xlarge",
        "ram": 65536
    },
    "c5ad.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.large",
            "instancesku": "HT8MSHND4S9PP8YT",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5ad.large",
        "name": "c5ad.large",
        "ram": 4096
    },
    "c5ad.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5ad.xlarge",
        "name": "c5ad.xlarge",
        "ram": 8192
    },
    "c5d.12xlarge": {
        "bandwidth": 12,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "188",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Deep Learning Boost; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5d.12xlarge",
        "name": "c5d.12xlarge",
        "ram": 98304
    },
    "c5d.18xlarge": {
        "bandwidth": 25,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "281",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.18xlarge",
            "instancesku": "RDHAU3QJUFDHWGAF",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "144 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "144",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "72",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5d.18xlarge",
        "name": "c5d.18xlarge",
        "ram": 147456
    },
    "c5d.24xlarge": {
        "bandwidth": 25,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "375",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.24xlarge",
            "instancesku": "CBCNASX9DSMBPMQW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Deep Learning Boost; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5d.24xlarge",
        "name": "c5d.24xlarge",
        "ram": 196608
    },
    "c5d.2xlarge": {
        "bandwidth": 10,
        "disk": 200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "39",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 200 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5d.2xlarge",
        "name": "c5d.2xlarge",
        "ram": 16384
    },
    "c5d.4xlarge": {
        "bandwidth": 10,
        "disk": 400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "73",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.4xlarge",
            "instancesku": "GZA4PSQEMRJ3BUYN",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 400 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5d.4xlarge",
        "name": "c5d.4xlarge",
        "ram": 32768
    },
    "c5d.9xlarge": {
        "bandwidth": 10,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "139",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.9xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "72 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "72",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 900 NVMe SSD",
            "vcpu": "36",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5d.9xlarge",
        "name": "c5d.9xlarge",
        "ram": 73728
    },
    "c5d.large": {
        "bandwidth": 10,
        "disk": 50,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 50 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5d.large",
        "name": "c5d.large",
        "ram": 4096
    },
    "c5d.xlarge": {
        "bandwidth": 10,
        "disk": 100,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "20",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.xlarge",
            "instancesku": "BEKRK57J4EHTR2VJ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 100 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5d.xlarge",
        "name": "c5d.xlarge",
        "ram": 8192
    },
    "c5n.18xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "281",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.18xlarge",
            "instancesku": "4VUXUHRQ7ZTHKVDG",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "144",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "72",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5n.18xlarge",
        "name": "c5n.18xlarge",
        "ram": 196608
    },
    "c5n.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "39",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "21 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5n.2xlarge",
        "name": "c5n.2xlarge",
        "ram": 21504
    },
    "c5n.4xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "73",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.4xlarge",
            "instancesku": "73YJN6HJETZG4DK2",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "42 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5n.4xlarge",
        "name": "c5n.4xlarge",
        "ram": 43008
    },
    "c5n.9xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "139",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.9xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "72",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "36",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5n.9xlarge",
        "name": "c5n.9xlarge",
        "ram": 98304
    },
    "c5n.large": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "5.25 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5n.large",
        "name": "c5n.large",
        "ram": 5376
    },
    "c5n.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "20",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "10.5 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "c5n.xlarge",
        "name": "c5n.xlarge",
        "ram": 10752
    },
    "c6g.12xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.12xlarge",
            "instancesku": "5427FJVXMN85SC8G",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6g.12xlarge",
        "name": "c6g.12xlarge",
        "ram": 98304
    },
    "c6g.16xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.16xlarge",
            "instancesku": "M8TJWAS9QQMZ7D8T",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6g.16xlarge",
        "name": "c6g.16xlarge",
        "ram": 131072
    },
    "c6g.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.2xlarge",
            "instancesku": "64UAZ5JQ6EFAVPMB",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6g.2xlarge",
        "name": "c6g.2xlarge",
        "ram": 16384
    },
    "c6g.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6g.4xlarge",
        "name": "c6g.4xlarge",
        "ram": 32768
    },
    "c6g.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.8xlarge",
            "instancesku": "PB5P6ZJYNWR2SRGE",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6g.8xlarge",
        "name": "c6g.8xlarge",
        "ram": 65536
    },
    "c6g.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.large",
            "instancesku": "6PQ4U29URPRQ2MYZ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6g.large",
        "name": "c6g.large",
        "ram": 4096
    },
    "c6g.medium": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6g.medium",
        "name": "c6g.medium",
        "ram": 2048
    },
    "c6g.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6g.xlarge",
        "name": "c6g.xlarge",
        "ram": 8192
    },
    "c6gd.12xlarge": {
        "bandwidth": 20,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gd.12xlarge",
        "name": "c6gd.12xlarge",
        "ram": 98304
    },
    "c6gd.16xlarge": {
        "bandwidth": 25,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "18000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gd.16xlarge",
        "name": "c6gd.16xlarge",
        "ram": 131072
    },
    "c6gd.2xlarge": {
        "bandwidth": 10,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.2xlarge",
            "instancesku": "VS68MNYP4R23S8R5",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gd.2xlarge",
        "name": "c6gd.2xlarge",
        "ram": 16384
    },
    "c6gd.4xlarge": {
        "bandwidth": 10,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.4xlarge",
            "instancesku": "8DYFS6J38M6GB9JE",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gd.4xlarge",
        "name": "c6gd.4xlarge",
        "ram": 32768
    },
    "c6gd.8xlarge": {
        "bandwidth": 10,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gd.8xlarge",
        "name": "c6gd.8xlarge",
        "ram": 65536
    },
    "c6gd.large": {
        "bandwidth": 10,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gd.large",
        "name": "c6gd.large",
        "ram": 4096
    },
    "c6gd.medium": {
        "bandwidth": 10,
        "disk": 59,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.medium",
            "instancesku": "FSJHC9R7RFGREHMN",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 59 NVMe SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gd.medium",
        "name": "c6gd.medium",
        "ram": 2048
    },
    "c6gd.xlarge": {
        "bandwidth": 10,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.xlarge",
            "instancesku": "8JNY7VUMPB55FZXV",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gd.xlarge",
        "name": "c6gd.xlarge",
        "ram": 8192
    },
    "c6gn.12xlarge": {
        "bandwidth": 75,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "28500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gn.12xlarge",
        "name": "c6gn.12xlarge",
        "ram": 98304
    },
    "c6gn.16xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gn.16xlarge",
        "name": "c6gn.16xlarge",
        "ram": 131072
    },
    "c6gn.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gn.2xlarge",
        "name": "c6gn.2xlarge",
        "ram": 16384
    },
    "c6gn.4xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gn.4xlarge",
        "name": "c6gn.4xlarge",
        "ram": 32768
    },
    "c6gn.8xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.8xlarge",
            "instancesku": "NX8CJEX2JY5HXHQR",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gn.8xlarge",
        "name": "c6gn.8xlarge",
        "ram": 65536
    },
    "c6gn.large": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gn.large",
        "name": "c6gn.large",
        "ram": 4096
    },
    "c6gn.medium": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gn.medium",
        "name": "c6gn.medium",
        "ram": 2048
    },
    "c6gn.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6gn.xlarge",
        "name": "c6gn.xlarge",
        "ram": 8192
    },
    "c6i.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6i.12xlarge",
        "name": "c6i.12xlarge",
        "ram": 98304
    },
    "c6i.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.16xlarge",
            "instancesku": "R289TCRFFZT3U2N2",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6i.16xlarge",
        "name": "c6i.16xlarge",
        "ram": 131072
    },
    "c6i.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.24xlarge",
            "instancesku": "ESDGN9RFWKFW97FE",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6i.24xlarge",
        "name": "c6i.24xlarge",
        "ram": 196608
    },
    "c6i.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.2xlarge",
            "instancesku": "DJ2YJFDQ4XVGDRF7",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6i.2xlarge",
        "name": "c6i.2xlarge",
        "ram": 16384
    },
    "c6i.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.32xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6i.32xlarge",
        "name": "c6i.32xlarge",
        "ram": 262144
    },
    "c6i.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.4xlarge",
            "instancesku": "4H74EXBA3BH2E77C",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6i.4xlarge",
        "name": "c6i.4xlarge",
        "ram": 32768
    },
    "c6i.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6i.8xlarge",
        "name": "c6i.8xlarge",
        "ram": 65536
    },
    "c6i.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6i.large",
        "name": "c6i.large",
        "ram": 4096
    },
    "c6i.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.xlarge",
            "instancesku": "TJG66TUBQUDBMGTP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "c6i.xlarge",
        "name": "c6i.xlarge",
        "ram": 8192
    },
    "cc2.8xlarge": {
        "bandwidth": 10,
        "disk": 3360,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "No",
            "ecu": "88",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Compute optimized",
            "instanceType": "cc2.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "60.5 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2670",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 840 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "cc2.8xlarge",
        "name": "cc2.8xlarge",
        "ram": 61952
    },
    "cr1.8xlarge": {
        "bandwidth": 10,
        "disk": 240,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "88",
            "instanceFamily": "Memory optimized",
            "instanceType": "cr1.8xlarge",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2670",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 120 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "cr1.8xlarge",
        "name": "cr1.8xlarge",
        "ram": 249856
    },
    "d2.2xlarge": {
        "bandwidth": None,
        "disk": 12000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1000 Mbps",
            "ecu": "28",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d2.2xlarge",
            "instancesku": "MMEKDXY58RD4JWKV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "6 x 2000 HDD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "d2.2xlarge",
        "name": "d2.2xlarge",
        "ram": 62464
    },
    "d2.4xlarge": {
        "bandwidth": None,
        "disk": 24000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2000 Mbps",
            "ecu": "56",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d2.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "12 x 2000 HDD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "d2.4xlarge",
        "name": "d2.4xlarge",
        "ram": 124928
    },
    "d2.8xlarge": {
        "bandwidth": 10,
        "disk": 48000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4000 Mbps",
            "ecu": "116",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d2.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "24 x 2000 HDD",
            "vcpu": "36",
            "vpcnetworkingsupport": "true"
        },
        "id": "d2.8xlarge",
        "name": "d2.8xlarge",
        "ram": 249856
    },
    "d2.xlarge": {
        "bandwidth": None,
        "disk": 6000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "750 Mbps",
            "ecu": "14",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d2.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "3 x 2000 HDD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "d2.xlarge",
        "name": "d2.xlarge",
        "ram": 31232
    },
    "d3.2xlarge": {
        "bandwidth": 15,
        "disk": 12000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3.2xlarge",
            "instancesku": "5TGGFNPG4M6GTUKU",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "6 x 2000 HDD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "d3.2xlarge",
        "name": "d3.2xlarge",
        "ram": 65536
    },
    "d3.4xlarge": {
        "bandwidth": 15,
        "disk": 24000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3.4xlarge",
            "instancesku": "4DVAYKJPKD2C77C6",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "12 x 2000 HDD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "d3.4xlarge",
        "name": "d3.4xlarge",
        "ram": 131072
    },
    "d3.8xlarge": {
        "bandwidth": 25,
        "disk": 48000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "24 x 2000 HDD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "d3.8xlarge",
        "name": "d3.8xlarge",
        "ram": 262144
    },
    "d3.xlarge": {
        "bandwidth": 15,
        "disk": 6000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3.xlarge",
            "instancesku": "E6KXGC3UNB34ACFA",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "3 x 2000 HDD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "d3.xlarge",
        "name": "d3.xlarge",
        "ram": 32768
    },
    "d3en.12xlarge": {
        "bandwidth": 75,
        "disk": 336000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.12xlarge",
            "instancesku": "2DB47P5DFHGJQV9E",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "24 x 14000 HDD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "d3en.12xlarge",
        "name": "d3en.12xlarge",
        "ram": 196608
    },
    "d3en.2xlarge": {
        "bandwidth": 25,
        "disk": 56000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.2xlarge",
            "instancesku": "PYH7JAKYH826PKNQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 14000 HDD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "d3en.2xlarge",
        "name": "d3en.2xlarge",
        "ram": 32768
    },
    "d3en.4xlarge": {
        "bandwidth": 25,
        "disk": 112000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.4xlarge",
            "instancesku": "V9KYSP3C455JHSEV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 14000 HDD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "d3en.4xlarge",
        "name": "d3en.4xlarge",
        "ram": 65536
    },
    "d3en.6xlarge": {
        "bandwidth": 40,
        "disk": 168000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.6xlarge",
            "instancesku": "CYYSUVKDC5DVFZCB",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "40 Gigabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "12 x 14000 HDD",
            "vcpu": "24",
            "vpcnetworkingsupport": "true"
        },
        "id": "d3en.6xlarge",
        "name": "d3en.6xlarge",
        "ram": 98304
    },
    "d3en.8xlarge": {
        "bandwidth": 50,
        "disk": 224000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.8xlarge",
            "instancesku": "KC4ZMWJZMF9KCPSJ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "16 x 14000 HDD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "d3en.8xlarge",
        "name": "d3en.8xlarge",
        "ram": 131072
    },
    "d3en.xlarge": {
        "bandwidth": 25,
        "disk": 28000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.xlarge",
            "instancesku": "9XAK7T9XR83GCZVV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 14000 HDD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "d3en.xlarge",
        "name": "d3en.xlarge",
        "ram": 16384
    },
    "dl1.24xlarge": {
        "bandwidth": 400,
        "disk": 4000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "dl1.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "400 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1000 GB NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "dl1.24xlarge",
        "name": "dl1.24xlarge",
        "ram": 786432
    },
    "f1.16xlarge": {
        "bandwidth": 20,
        "disk": 3760,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "FPGA Instances",
            "instanceType": "f1.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "976 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 940 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "f1.16xlarge",
        "name": "f1.16xlarge",
        "ram": 999424
    },
    "f1.2xlarge": {
        "bandwidth": 10,
        "disk": 470,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1700 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "FPGA Instances",
            "instanceType": "f1.2xlarge",
            "instancesku": "MXDPYTEVHNYM6QTA",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 470 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "f1.2xlarge",
        "name": "f1.2xlarge",
        "ram": 124928
    },
    "f1.4xlarge": {
        "bandwidth": 10,
        "disk": 940,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "58",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "FPGA Instances",
            "instanceType": "f1.4xlarge",
            "instancesku": "6YUJAEPDDF47NZUQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 940 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "f1.4xlarge",
        "name": "f1.4xlarge",
        "ram": 249856
    },
    "g2.2xlarge": {
        "bandwidth": None,
        "disk": 60,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "No",
            "ecu": "26",
            "enhancedNetworkingSupported": "No",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "g2.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2670 (Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 60 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "g2.2xlarge",
        "name": "g2.2xlarge",
        "ram": 15360
    },
    "g2.8xlarge": {
        "bandwidth": None,
        "disk": 240,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "No",
            "ecu": "104",
            "enhancedNetworkingSupported": "No",
            "gpu": "4",
            "instanceFamily": "GPU instance",
            "instanceType": "g2.8xlarge",
            "instancesku": "GP54RNQ37B9N43MJ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "60 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2670 (Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 120 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "g2.8xlarge",
        "name": "g2.8xlarge",
        "ram": 61440
    },
    "g3.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "instanceFamily": "GPU instance",
            "instanceType": "g3.16xlarge",
            "instancesku": "9JX6W8YKTGAXCFNB",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "g3.16xlarge",
        "name": "g3.16xlarge",
        "ram": 499712
    },
    "g3.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "58",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "g3.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "g3.4xlarge",
        "name": "g3.4xlarge",
        "ram": 124928
    },
    "g3.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "2",
            "instanceFamily": "GPU instance",
            "instanceType": "g3.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "g3.8xlarge",
        "name": "g3.8xlarge",
        "ram": 249856
    },
    "g3s.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "100 Mbps",
            "ecu": "13",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "g3s.xlarge",
            "instancesku": "H33F2YFNF92GTBPB",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "g3s.xlarge",
        "name": "g3s.xlarge",
        "ram": 31232
    },
    "g4ad.16xlarge": {
        "bandwidth": 25,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "instanceFamily": "GPU instance",
            "instanceType": "g4ad.16xlarge",
            "instancesku": "UTWC8VBRTBSXJYDB",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2400 GB NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "g4ad.16xlarge",
        "name": "g4ad.16xlarge",
        "ram": 262144
    },
    "g4ad.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "g4ad.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "19.97394",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "300 GB NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "g4ad.2xlarge",
        "name": "g4ad.2xlarge",
        "ram": 32768
    },
    "g4ad.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "g4ad.4xlarge",
            "instancesku": "9VMBVH3KEV82T8ZW",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "600 GB NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "g4ad.4xlarge",
        "name": "g4ad.4xlarge",
        "ram": 65536
    },
    "g4ad.8xlarge": {
        "bandwidth": 15,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "2",
            "instanceFamily": "GPU instance",
            "instanceType": "g4ad.8xlarge",
            "instancesku": "D6FUGEGN8P4Q7E97",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1200 GB NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "g4ad.8xlarge",
        "name": "g4ad.8xlarge",
        "ram": 131072
    },
    "g4ad.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "g4ad.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "13.97113",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "150 GB NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "g4ad.xlarge",
        "name": "g4ad.xlarge",
        "ram": 16384
    },
    "g4dn.12xlarge": {
        "bandwidth": 50,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.12xlarge",
            "instancesku": "5GY7VGBKXNCWB5JE",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "900 GB NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "g4dn.12xlarge",
        "name": "g4dn.12xlarge",
        "ram": 196608
    },
    "g4dn.16xlarge": {
        "bandwidth": 50,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "900 GB NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "g4dn.16xlarge",
        "name": "g4dn.16xlarge",
        "ram": 262144
    },
    "g4dn.2xlarge": {
        "bandwidth": 25,
        "disk": 225,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.2xlarge",
            "instancesku": "K24CAS4K6W4HJXN8",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "225 GB NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "g4dn.2xlarge",
        "name": "g4dn.2xlarge",
        "ram": 32768
    },
    "g4dn.4xlarge": {
        "bandwidth": 25,
        "disk": 225,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.4xlarge",
            "instancesku": "YMZGCYKDK8626BJT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "225 GB NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "g4dn.4xlarge",
        "name": "g4dn.4xlarge",
        "ram": 65536
    },
    "g4dn.8xlarge": {
        "bandwidth": 50,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.8xlarge",
            "instancesku": "AE5GUKCAC8GSBSBD",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "900 GB NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "g4dn.8xlarge",
        "name": "g4dn.8xlarge",
        "ram": 131072
    },
    "g4dn.xlarge": {
        "bandwidth": 25,
        "disk": 125,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.xlarge",
            "instancesku": "V5NBPKWB9EBYQAHY",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "125 GB NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "g4dn.xlarge",
        "name": "g4dn.xlarge",
        "ram": 16384
    },
    "h1.16xlarge": {
        "bandwidth": 25,
        "disk": 16000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "h1.16xlarge",
            "instancesku": "42A6HHN38X5E4VKA",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 2000 HDD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "h1.16xlarge",
        "name": "h1.16xlarge",
        "ram": 262144
    },
    "h1.2xlarge": {
        "bandwidth": 10,
        "disk": 2000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1750 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "h1.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 2000 HDD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "h1.2xlarge",
        "name": "h1.2xlarge",
        "ram": 32768
    },
    "h1.4xlarge": {
        "bandwidth": 10,
        "disk": 4000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "58",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "h1.4xlarge",
            "instancesku": "NGJJD4N26GD3FZRX",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 2000 HDD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "h1.4xlarge",
        "name": "h1.4xlarge",
        "ram": 65536
    },
    "h1.8xlarge": {
        "bandwidth": 10,
        "disk": 8000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "h1.8xlarge",
            "instancesku": "WH74JE96HBFW3TTP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 2000 HDD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "h1.8xlarge",
        "name": "h1.8xlarge",
        "ram": 131072
    },
    "hs1.8xlarge": {
        "bandwidth": 10,
        "disk": 48000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2 GHz",
            "currentGeneration": "No",
            "ecu": "35",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Storage optimized",
            "instanceType": "hs1.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "117 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2650",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "24 x 2000 HDD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "hs1.8xlarge",
        "name": "hs1.8xlarge",
        "ram": 119808
    },
    "i2.2xlarge": {
        "bandwidth": None,
        "disk": 1600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "1500 Mbps",
            "ecu": "27",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i2.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 800 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "i2.2xlarge",
        "name": "i2.2xlarge",
        "ram": 62464
    },
    "i2.4xlarge": {
        "bandwidth": None,
        "disk": 3200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "1500 Mbps",
            "ecu": "53",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i2.4xlarge",
            "instancesku": "Y9Z5DVR4T6JJHPDH",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 800 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "i2.4xlarge",
        "name": "i2.4xlarge",
        "ram": 124928
    },
    "i2.8xlarge": {
        "bandwidth": 10,
        "disk": 6400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "1500 Mbps",
            "ecu": "104",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i2.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 800 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "i2.8xlarge",
        "name": "i2.8xlarge",
        "ram": 249856
    },
    "i2.large": {
        "bandwidth": None,
        "disk": 800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "7",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i2.large",
            "instancesku": "DNMJX39QXYR342B4",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 800 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "i2.large",
        "name": "i2.large",
        "ram": 15360
    },
    "i2.xlarge": {
        "bandwidth": None,
        "disk": 800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "750 Mbps",
            "ecu": "14",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i2.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 800 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "i2.xlarge",
        "name": "i2.xlarge",
        "ram": 31232
    },
    "i3.16xlarge": {
        "bandwidth": 20,
        "disk": 15200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3.16xlarge",
        "name": "i3.16xlarge",
        "ram": 499712
    },
    "i3.2xlarge": {
        "bandwidth": 10,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1750 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3.2xlarge",
        "name": "i3.2xlarge",
        "ram": 62464
    },
    "i3.4xlarge": {
        "bandwidth": 10,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "58",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.4xlarge",
            "instancesku": "A4Y3Z2HDYDB9Y278",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3.4xlarge",
        "name": "i3.4xlarge",
        "ram": 124928
    },
    "i3.8xlarge": {
        "bandwidth": 10,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3.8xlarge",
        "name": "i3.8xlarge",
        "ram": 249856
    },
    "i3.large": {
        "bandwidth": 10,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "425 Mbps",
            "ecu": "8",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15.25 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3.large",
        "name": "i3.large",
        "ram": 15616
    },
    "i3.xlarge": {
        "bandwidth": 10,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "850 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.xlarge",
            "instancesku": "DQBUA3WF2ZKB94XE",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3.xlarge",
        "name": "i3.xlarge",
        "ram": 31232
    },
    "i3en.12xlarge": {
        "bandwidth": 50,
        "disk": 30000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "168",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.12xlarge",
            "instancesku": "9JQFRDVY79XEF6QA",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 7500 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3en.12xlarge",
        "name": "i3en.12xlarge",
        "ram": 393216
    },
    "i3en.24xlarge": {
        "bandwidth": 100,
        "disk": 60000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 7500 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3en.24xlarge",
        "name": "i3en.24xlarge",
        "ram": 786432
    },
    "i3en.2xlarge": {
        "bandwidth": 25,
        "disk": 5000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "37",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 2500 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3en.2xlarge",
        "name": "i3en.2xlarge",
        "ram": 65536
    },
    "i3en.3xlarge": {
        "bandwidth": 25,
        "disk": 7500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.3xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "24",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 7500 NVMe SSD",
            "vcpu": "12",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3en.3xlarge",
        "name": "i3en.3xlarge",
        "ram": 98304
    },
    "i3en.6xlarge": {
        "bandwidth": 25,
        "disk": 15000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.6xlarge",
            "instancesku": "JFUSUE6PEJ7BEWFW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 7500 NVMe SSD",
            "vcpu": "24",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3en.6xlarge",
        "name": "i3en.6xlarge",
        "ram": 196608
    },
    "i3en.large": {
        "bandwidth": 25,
        "disk": 1250,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.large",
            "instancesku": "9KJJPTN45B7DSGAW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1250 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3en.large",
        "name": "i3en.large",
        "ram": 16384
    },
    "i3en.xlarge": {
        "bandwidth": 25,
        "disk": 2500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 2500 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "i3en.xlarge",
        "name": "i3en.xlarge",
        "ram": 32768
    },
    "inf1.24xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf1.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Platinum 8275CL (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "inf1.24xlarge",
        "name": "inf1.24xlarge",
        "ram": 196608
    },
    "inf1.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "875 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf1.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Platinum 8275CL (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "inf1.2xlarge",
        "name": "inf1.2xlarge",
        "ram": 16384
    },
    "inf1.6xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf1.6xlarge",
            "instancesku": "WFATCAUXJSST4KZ8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "48 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Platinum 8275CL (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "24",
            "vpcnetworkingsupport": "true"
        },
        "id": "inf1.6xlarge",
        "name": "inf1.6xlarge",
        "ram": 49152
    },
    "inf1.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "875 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf1.xlarge",
            "instancesku": "26TMKJU38MB7PCPQ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Platinum 8275CL (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "inf1.xlarge",
        "name": "inf1.xlarge",
        "ram": 8192
    },
    "m1.large": {
        "bandwidth": None,
        "disk": 840,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "4",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m1.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "7.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 420 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m1.large",
        "name": "m1.large",
        "ram": 7680
    },
    "m1.medium": {
        "bandwidth": None,
        "disk": 410,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "2",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m1.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "3.75 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 410 SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "m1.medium",
        "name": "m1.medium",
        "ram": 3840
    },
    "m1.small": {
        "bandwidth": None,
        "disk": 160,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "1",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m1.small",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1.7 GiB",
            "networkPerformance": "Low",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 160 SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "m1.small",
        "name": "m1.small",
        "ram": 1740
    },
    "m1.xlarge": {
        "bandwidth": None,
        "disk": 1680,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "8",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m1.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 420 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m1.xlarge",
        "name": "m1.xlarge",
        "ram": 15360
    },
    "m2.2xlarge": {
        "bandwidth": None,
        "disk": 850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "13",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "m2.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "34.2 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 850 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m2.2xlarge",
        "name": "m2.2xlarge",
        "ram": 35020
    },
    "m2.4xlarge": {
        "bandwidth": None,
        "disk": 1680,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "26",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "m2.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "68.4 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 840 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m2.4xlarge",
        "name": "m2.4xlarge",
        "ram": 70041
    },
    "m2.xlarge": {
        "bandwidth": None,
        "disk": 420,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "6.5",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "m2.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "17.1 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 420 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m2.xlarge",
        "name": "m2.xlarge",
        "ram": 17510
    },
    "m3.2xlarge": {
        "bandwidth": None,
        "disk": 160,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "1000 Mbps",
            "ecu": "26",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m3.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge/Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 80 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m3.2xlarge",
        "name": "m3.2xlarge",
        "ram": 30720
    },
    "m3.large": {
        "bandwidth": None,
        "disk": 32,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "6.5",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m3.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "7.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge/Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 32 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m3.large",
        "name": "m3.large",
        "ram": 7680
    },
    "m3.medium": {
        "bandwidth": None,
        "disk": 4,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "3",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m3.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "3.75 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge/Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 4 SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "m3.medium",
        "name": "m3.medium",
        "ram": 3840
    },
    "m3.xlarge": {
        "bandwidth": None,
        "disk": 80,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "500 Mbps",
            "ecu": "13",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m3.xlarge",
            "instancesku": "JCFYMXQEQ3BA57K3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge/Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 40 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m3.xlarge",
        "name": "m3.xlarge",
        "ram": 15360
    },
    "m4.10xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4000 Mbps",
            "ecu": "124.5",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m4.10xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "160 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "80",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "40",
            "vpcnetworkingsupport": "true"
        },
        "id": "m4.10xlarge",
        "name": "m4.10xlarge",
        "ram": 163840
    },
    "m4.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "188",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m4.16xlarge",
            "instancesku": "9RRW6TTJ2K77NSDQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "m4.16xlarge",
        "name": "m4.16xlarge",
        "ram": 262144
    },
    "m4.2xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1000 Mbps",
            "ecu": "26",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m4.2xlarge",
            "instancesku": "4DP9JW528SDT7HRV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m4.2xlarge",
        "name": "m4.2xlarge",
        "ram": 32768
    },
    "m4.4xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2000 Mbps",
            "ecu": "53.5",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m4.4xlarge",
            "instancesku": "B5A2KQPVUB48X3ES",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "m4.4xlarge",
        "name": "m4.4xlarge",
        "ram": 65536
    },
    "m4.large": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "450 Mbps",
            "ecu": "6.5",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m4.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m4.large",
        "name": "m4.large",
        "ram": 8192
    },
    "m4.xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "750 Mbps",
            "ecu": "13",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m4.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m4.xlarge",
        "name": "m4.xlarge",
        "ram": 16384
    },
    "m5.12xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "168",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5.12xlarge",
            "instancesku": "P4TMXGTTR2DKGBTH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5.12xlarge",
        "name": "m5.12xlarge",
        "ram": 196608
    },
    "m5.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "256",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5.16xlarge",
        "name": "m5.16xlarge",
        "ram": 262144
    },
    "m5.24xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5.24xlarge",
            "instancesku": "SCEVK3MMHE5VFGJQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5.24xlarge",
        "name": "m5.24xlarge",
        "ram": 393216
    },
    "m5.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "37",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5.2xlarge",
            "instancesku": "7DHN2N5PWU2GRR8X",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5.2xlarge",
        "name": "m5.2xlarge",
        "ram": 32768
    },
    "m5.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "70",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5.4xlarge",
        "name": "m5.4xlarge",
        "ram": 65536
    },
    "m5.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "128",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5.8xlarge",
            "instancesku": "ATEPRR4KUQNVY664",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5.8xlarge",
        "name": "m5.8xlarge",
        "ram": 131072
    },
    "m5.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5.large",
            "instancesku": "82KMZB2MXBZ899J2",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5.large",
        "name": "m5.large",
        "ram": 8192
    },
    "m5.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5.xlarge",
        "name": "m5.xlarge",
        "ram": 16384
    },
    "m5a.12xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5a.12xlarge",
        "name": "m5a.12xlarge",
        "ram": 196608
    },
    "m5a.16xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.16xlarge",
            "instancesku": "GXMYJ4VW7W2QGTWP",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5a.16xlarge",
        "name": "m5a.16xlarge",
        "ram": 262144
    },
    "m5a.24xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5a.24xlarge",
        "name": "m5a.24xlarge",
        "ram": 393216
    },
    "m5a.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.2xlarge",
            "instancesku": "NP557QB7VFTC48QX",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5a.2xlarge",
        "name": "m5a.2xlarge",
        "ram": 32768
    },
    "m5a.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.4xlarge",
            "instancesku": "YRM7GSGQEF9HKS2Q",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5a.4xlarge",
        "name": "m5a.4xlarge",
        "ram": 65536
    },
    "m5a.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5a.8xlarge",
        "name": "m5a.8xlarge",
        "ram": 131072
    },
    "m5a.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.large",
            "instancesku": "F226W5UEGVDUWR26",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5a.large",
        "name": "m5a.large",
        "ram": 8192
    },
    "m5a.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.xlarge",
            "instancesku": "8XHTUYUX8E6EQX3E",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5a.xlarge",
        "name": "m5a.xlarge",
        "ram": 16384
    },
    "m5ad.12xlarge": {
        "bandwidth": 10,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5ad.12xlarge",
        "name": "m5ad.12xlarge",
        "ram": 196608
    },
    "m5ad.16xlarge": {
        "bandwidth": 12,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5ad.16xlarge",
        "name": "m5ad.16xlarge",
        "ram": 262144
    },
    "m5ad.24xlarge": {
        "bandwidth": 20,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.24xlarge",
            "instancesku": "R24J62ZT83FRAWE4",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5ad.24xlarge",
        "name": "m5ad.24xlarge",
        "ram": 393216
    },
    "m5ad.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.2xlarge",
            "instancesku": "NC7JAPBAQR9N6AKC",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5ad.2xlarge",
        "name": "m5ad.2xlarge",
        "ram": 32768
    },
    "m5ad.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.4xlarge",
            "instancesku": "B42N5BR67T9VQCUZ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5ad.4xlarge",
        "name": "m5ad.4xlarge",
        "ram": 65536
    },
    "m5ad.8xlarge": {
        "bandwidth": 10,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5ad.8xlarge",
        "name": "m5ad.8xlarge",
        "ram": 131072
    },
    "m5ad.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5ad.large",
        "name": "m5ad.large",
        "ram": 8192
    },
    "m5ad.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.xlarge",
            "instancesku": "746NNG5CG6EEY9BJ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5ad.xlarge",
        "name": "m5ad.xlarge",
        "ram": 16384
    },
    "m5d.12xlarge": {
        "bandwidth": 10,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "168",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.12xlarge",
            "instancesku": "UBWKCRY37QGF892F",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5d.12xlarge",
        "name": "m5d.12xlarge",
        "ram": 196608
    },
    "m5d.16xlarge": {
        "bandwidth": 20,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "256",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.16xlarge",
            "instancesku": "2AW5G4CMNCDS44AT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5d.16xlarge",
        "name": "m5d.16xlarge",
        "ram": 262144
    },
    "m5d.24xlarge": {
        "bandwidth": 25,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5d.24xlarge",
        "name": "m5d.24xlarge",
        "ram": 393216
    },
    "m5d.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "37",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5d.2xlarge",
        "name": "m5d.2xlarge",
        "ram": 32768
    },
    "m5d.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "70",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.4xlarge",
            "instancesku": "XHECG69V2ZSYT8UV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5d.4xlarge",
        "name": "m5d.4xlarge",
        "ram": 65536
    },
    "m5d.8xlarge": {
        "bandwidth": 10,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "128",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.8xlarge",
            "instancesku": "2C2YGPB7B4MPSHV7",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5d.8xlarge",
        "name": "m5d.8xlarge",
        "ram": 131072
    },
    "m5d.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.large",
            "instancesku": "GJE7QRBNDAYYEUHG",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5d.large",
        "name": "m5d.large",
        "ram": 8192
    },
    "m5d.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.xlarge",
            "instancesku": "GKEVTCW4GXTX52AW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5d.xlarge",
        "name": "m5d.xlarge",
        "ram": 16384
    },
    "m5dn.12xlarge": {
        "bandwidth": 50,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 GB NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5dn.12xlarge",
        "name": "m5dn.12xlarge",
        "ram": 196608
    },
    "m5dn.16xlarge": {
        "bandwidth": 75,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5dn.16xlarge",
        "name": "m5dn.16xlarge",
        "ram": 262144
    },
    "m5dn.24xlarge": {
        "bandwidth": 100,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.24xlarge",
            "instancesku": "G9H536RMY737GEM4",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5dn.24xlarge",
        "name": "m5dn.24xlarge",
        "ram": 393216
    },
    "m5dn.2xlarge": {
        "bandwidth": 25,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.2xlarge",
            "instancesku": "3WZA3VEJ8QD5UJBE",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5dn.2xlarge",
        "name": "m5dn.2xlarge",
        "ram": 32768
    },
    "m5dn.4xlarge": {
        "bandwidth": 25,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.4xlarge",
            "instancesku": "3Y5Q72YU4JNPU6S5",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5dn.4xlarge",
        "name": "m5dn.4xlarge",
        "ram": 65536
    },
    "m5dn.8xlarge": {
        "bandwidth": 25,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.8xlarge",
            "instancesku": "N53VJWC86E4NYNWY",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5dn.8xlarge",
        "name": "m5dn.8xlarge",
        "ram": 131072
    },
    "m5dn.large": {
        "bandwidth": 25,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.large",
            "instancesku": "SVUJNCR3NN2359RD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5dn.large",
        "name": "m5dn.large",
        "ram": 8192
    },
    "m5dn.xlarge": {
        "bandwidth": 25,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5dn.xlarge",
        "name": "m5dn.xlarge",
        "ram": 16384
    },
    "m5n.12xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.12xlarge",
            "instancesku": "BEXX757YG98NCW3T",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5n.12xlarge",
        "name": "m5n.12xlarge",
        "ram": 196608
    },
    "m5n.16xlarge": {
        "bandwidth": 75,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.16xlarge",
            "instancesku": "88HNSFX9MCY6YAC3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5n.16xlarge",
        "name": "m5n.16xlarge",
        "ram": 262144
    },
    "m5n.24xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5n.24xlarge",
        "name": "m5n.24xlarge",
        "ram": 393216
    },
    "m5n.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5n.2xlarge",
        "name": "m5n.2xlarge",
        "ram": 32768
    },
    "m5n.4xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.4xlarge",
            "instancesku": "ZH4K42CV65578MRJ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5n.4xlarge",
        "name": "m5n.4xlarge",
        "ram": 65536
    },
    "m5n.8xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.8xlarge",
            "instancesku": "B9H9FBY9ZKWXQT4E",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5n.8xlarge",
        "name": "m5n.8xlarge",
        "ram": 131072
    },
    "m5n.large": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5n.large",
        "name": "m5n.large",
        "ram": 8192
    },
    "m5n.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5n.xlarge",
        "name": "m5n.xlarge",
        "ram": 16384
    },
    "m5zn.12xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.12xlarge",
            "instancesku": "USMQJ4AAYGRVZ5AP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5zn.12xlarge",
        "name": "m5zn.12xlarge",
        "ram": 196608
    },
    "m5zn.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.2xlarge",
            "instancesku": "AKY928WUU3BEMP5B",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5zn.2xlarge",
        "name": "m5zn.2xlarge",
        "ram": 32768
    },
    "m5zn.3xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.3xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "48 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "24",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "12",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5zn.3xlarge",
        "name": "m5zn.3xlarge",
        "ram": 49152
    },
    "m5zn.6xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.6xlarge",
            "instancesku": "UKFGCB9AHS4FY7V6",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "24",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5zn.6xlarge",
        "name": "m5zn.6xlarge",
        "ram": 98304
    },
    "m5zn.large": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5zn.large",
        "name": "m5zn.large",
        "ram": 8192
    },
    "m5zn.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.xlarge",
            "instancesku": "Z5KETQXBVZT4M8PQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m5zn.xlarge",
        "name": "m5zn.xlarge",
        "ram": 16384
    },
    "m6g.12xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.12xlarge",
            "instancesku": "QD2WPXWQ5TSZ32UD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6g.12xlarge",
        "name": "m6g.12xlarge",
        "ram": 196608
    },
    "m6g.16xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "18000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6g.16xlarge",
        "name": "m6g.16xlarge",
        "ram": 262144
    },
    "m6g.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2250 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.2xlarge",
            "instancesku": "SX3J4P5RPTH53ZTN",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6g.2xlarge",
        "name": "m6g.2xlarge",
        "ram": 32768
    },
    "m6g.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6g.4xlarge",
        "name": "m6g.4xlarge",
        "ram": 65536
    },
    "m6g.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6g.8xlarge",
        "name": "m6g.8xlarge",
        "ram": 131072
    },
    "m6g.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "600 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6g.large",
        "name": "m6g.large",
        "ram": 8192
    },
    "m6g.medium": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "300 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6g.medium",
        "name": "m6g.medium",
        "ram": 4096
    },
    "m6g.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1125 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.xlarge",
            "instancesku": "UT993RDVVNCCZUQ8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6g.xlarge",
        "name": "m6g.xlarge",
        "ram": 16384
    },
    "m6gd.12xlarge": {
        "bandwidth": 20,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6gd.12xlarge",
        "name": "m6gd.12xlarge",
        "ram": 196608
    },
    "m6gd.16xlarge": {
        "bandwidth": 25,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "18000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6gd.16xlarge",
        "name": "m6gd.16xlarge",
        "ram": 262144
    },
    "m6gd.2xlarge": {
        "bandwidth": 10,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.2xlarge",
            "instancesku": "MEPDU66RSS2VZTBP",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6gd.2xlarge",
        "name": "m6gd.2xlarge",
        "ram": 32768
    },
    "m6gd.4xlarge": {
        "bandwidth": 10,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.4xlarge",
            "instancesku": "KPBAA5MV8872F65A",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6gd.4xlarge",
        "name": "m6gd.4xlarge",
        "ram": 65536
    },
    "m6gd.8xlarge": {
        "bandwidth": 10,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.8xlarge",
            "instancesku": "HKW2CNJU63NTP3VD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6gd.8xlarge",
        "name": "m6gd.8xlarge",
        "ram": 131072
    },
    "m6gd.large": {
        "bandwidth": 10,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.large",
            "instancesku": "499NZC7Z2KYZC8MF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6gd.large",
        "name": "m6gd.large",
        "ram": 8192
    },
    "m6gd.medium": {
        "bandwidth": 10,
        "disk": 59,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.medium",
            "instancesku": "F82GDBMZZ9YEY6AD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 59 NVMe SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6gd.medium",
        "name": "m6gd.medium",
        "ram": 4096
    },
    "m6gd.xlarge": {
        "bandwidth": 10,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.xlarge",
            "instancesku": "3F2Z3GCEDNF62F7T",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6gd.xlarge",
        "name": "m6gd.xlarge",
        "ram": 16384
    },
    "m6i.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.12xlarge",
            "instancesku": "XWT95UWVSUQJ3HFY",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6i.12xlarge",
        "name": "m6i.12xlarge",
        "ram": 196608
    },
    "m6i.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6i.16xlarge",
        "name": "m6i.16xlarge",
        "ram": 262144
    },
    "m6i.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6i.24xlarge",
        "name": "m6i.24xlarge",
        "ram": 393216
    },
    "m6i.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6i.2xlarge",
        "name": "m6i.2xlarge",
        "ram": 32768
    },
    "m6i.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.32xlarge",
            "instancesku": "J2589GXE2X4QGE87",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6i.32xlarge",
        "name": "m6i.32xlarge",
        "ram": 524288
    },
    "m6i.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6i.4xlarge",
        "name": "m6i.4xlarge",
        "ram": 65536
    },
    "m6i.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6i.8xlarge",
        "name": "m6i.8xlarge",
        "ram": 131072
    },
    "m6i.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6i.large",
        "name": "m6i.large",
        "ram": 8192
    },
    "m6i.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "m6i.xlarge",
        "name": "m6i.xlarge",
        "ram": 16384
    },
    "p2.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "16",
            "instanceFamily": "GPU instance",
            "instanceType": "p2.16xlarge",
            "instancesku": "8NZJVJPUUPFQD5XT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "732 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "p2.16xlarge",
        "name": "p2.16xlarge",
        "ram": 749568
    },
    "p2.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "instanceFamily": "GPU instance",
            "instanceType": "p2.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "p2.8xlarge",
        "name": "p2.8xlarge",
        "ram": 499712
    },
    "p2.xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "750 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "p2.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "p2.xlarge",
        "name": "p2.xlarge",
        "ram": 62464
    },
    "p3.16xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "instanceFamily": "GPU instance",
            "instanceType": "p3.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "p3.16xlarge",
        "name": "p3.16xlarge",
        "ram": 499712
    },
    "p3.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1750 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "instanceFamily": "GPU instance",
            "instanceType": "p3.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "p3.2xlarge",
        "name": "p3.2xlarge",
        "ram": 62464
    },
    "p3.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "instanceFamily": "GPU instance",
            "instanceType": "p3.8xlarge",
            "instancesku": "HCAHXPK7A9MBBZGN",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "p3.8xlarge",
        "name": "p3.8xlarge",
        "ram": 249856
    },
    "p3dn.24xlarge": {
        "bandwidth": 100,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "instanceFamily": "GPU instance",
            "instanceType": "p3dn.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "p3dn.24xlarge",
        "name": "p3dn.24xlarge",
        "ram": 786432
    },
    "p4d.24xlarge": {
        "bandwidth": 400,
        "disk": 8000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "345",
            "enhancedNetworkingSupported": "No",
            "gpu": "8",
            "instanceFamily": "GPU instance",
            "instanceType": "p4d.24xlarge",
            "instancesku": "CASSTWDHBMWC5PPU",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1152 GiB",
            "networkPerformance": "400 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 1000 SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "p4d.24xlarge",
        "name": "p4d.24xlarge",
        "ram": 1179648
    },
    "r3.2xlarge": {
        "bandwidth": None,
        "disk": 160,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "26",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r3.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 160 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "r3.2xlarge",
        "name": "r3.2xlarge",
        "ram": 62464
    },
    "r3.4xlarge": {
        "bandwidth": None,
        "disk": 320,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "52",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r3.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 320 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "r3.4xlarge",
        "name": "r3.4xlarge",
        "ram": 124928
    },
    "r3.8xlarge": {
        "bandwidth": 10,
        "disk": 640,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "104",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r3.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 320 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "r3.8xlarge",
        "name": "r3.8xlarge",
        "ram": 249856
    },
    "r3.large": {
        "bandwidth": None,
        "disk": 32,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "6.5",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r3.large",
            "instancesku": "JXRNUMXAZNYRPC72",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15.25 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 32 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "r3.large",
        "name": "r3.large",
        "ram": 15616
    },
    "r3.xlarge": {
        "bandwidth": None,
        "disk": 80,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "13",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r3.xlarge",
            "instancesku": "XKREJETCK4Q363EE",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 80 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "r3.xlarge",
        "name": "r3.xlarge",
        "ram": 31232
    },
    "r4.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "r4.16xlarge",
        "name": "r4.16xlarge",
        "ram": 499712
    },
    "r4.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1600 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "r4.2xlarge",
        "name": "r4.2xlarge",
        "ram": 62464
    },
    "r4.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3000 Mbps",
            "ecu": "58",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.4xlarge",
            "instancesku": "C9S5QU5S5T9EP9E3",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "r4.4xlarge",
        "name": "r4.4xlarge",
        "ram": 124928
    },
    "r4.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "r4.8xlarge",
        "name": "r4.8xlarge",
        "ram": 249856
    },
    "r4.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "400 Mbps",
            "ecu": "8",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.large",
            "instancesku": "CDTBMEVPWD5M6R8F",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15.25 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "r4.large",
        "name": "r4.large",
        "ram": 15616
    },
    "r4.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "800 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.xlarge",
            "instancesku": "YTUYTGV9WJQD2NAN",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "r4.xlarge",
        "name": "r4.xlarge",
        "ram": 31232
    },
    "r5.12xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "168",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5.12xlarge",
        "name": "r5.12xlarge",
        "ram": 393216
    },
    "r5.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13600 Mbps",
            "ecu": "256",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.16xlarge",
            "instancesku": "Z2E8HZWVPEUN65QE",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5.16xlarge",
        "name": "r5.16xlarge",
        "ram": 524288
    },
    "r5.24xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5.24xlarge",
        "name": "r5.24xlarge",
        "ram": 786432
    },
    "r5.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "37",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.2xlarge",
            "instancesku": "D5AEEW57JF2DX3U9",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5.2xlarge",
        "name": "r5.2xlarge",
        "ram": 65536
    },
    "r5.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "70",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5.4xlarge",
        "name": "r5.4xlarge",
        "ram": 131072
    },
    "r5.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6800 Mbps",
            "ecu": "128",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.8xlarge",
            "instancesku": "XF56BWYP9FKPXWCE",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5.8xlarge",
        "name": "r5.8xlarge",
        "ram": 262144
    },
    "r5.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5.large",
        "name": "r5.large",
        "ram": 16384
    },
    "r5.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "19",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.xlarge",
            "instancesku": "RCHCT9D9BT5VD66A",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5.xlarge",
        "name": "r5.xlarge",
        "ram": 32768
    },
    "r5a.12xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5a.12xlarge",
        "name": "r5a.12xlarge",
        "ram": 393216
    },
    "r5a.16xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.16xlarge",
            "instancesku": "2B3VAYDSHV6M39S6",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5a.16xlarge",
        "name": "r5a.16xlarge",
        "ram": 524288
    },
    "r5a.24xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.24xlarge",
            "instancesku": "YG2S4S8SBWEBG9X7",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5a.24xlarge",
        "name": "r5a.24xlarge",
        "ram": 786432
    },
    "r5a.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5a.2xlarge",
        "name": "r5a.2xlarge",
        "ram": 65536
    },
    "r5a.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.4xlarge",
            "instancesku": "9USH93WEV7YHKN3N",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5a.4xlarge",
        "name": "r5a.4xlarge",
        "ram": 131072
    },
    "r5a.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5a.8xlarge",
        "name": "r5a.8xlarge",
        "ram": 262144
    },
    "r5a.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.large",
            "instancesku": "HXFJ8J7KRRDCUQDF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5a.large",
        "name": "r5a.large",
        "ram": 16384
    },
    "r5a.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5a.xlarge",
        "name": "r5a.xlarge",
        "ram": 32768
    },
    "r5ad.12xlarge": {
        "bandwidth": 10,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.12xlarge",
            "instancesku": "BPGWTSU35274F5JD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5ad.12xlarge",
        "name": "r5ad.12xlarge",
        "ram": 393216
    },
    "r5ad.16xlarge": {
        "bandwidth": 12,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.16xlarge",
            "instancesku": "H8NDT2T8J8KSRGRD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5ad.16xlarge",
        "name": "r5ad.16xlarge",
        "ram": 524288
    },
    "r5ad.24xlarge": {
        "bandwidth": 20,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.24xlarge",
            "instancesku": "AYARQMMWY3KRT3Z8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5ad.24xlarge",
        "name": "r5ad.24xlarge",
        "ram": 786432
    },
    "r5ad.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.2xlarge",
            "instancesku": "ZR4V2JBYXNCCZXR3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5ad.2xlarge",
        "name": "r5ad.2xlarge",
        "ram": 65536
    },
    "r5ad.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5ad.4xlarge",
        "name": "r5ad.4xlarge",
        "ram": 131072
    },
    "r5ad.8xlarge": {
        "bandwidth": 10,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.8xlarge",
            "instancesku": "KKTZFU3YAWTUZQEF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5ad.8xlarge",
        "name": "r5ad.8xlarge",
        "ram": 262144
    },
    "r5ad.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.large",
            "instancesku": "NK72MQAHKR4TXF5P",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5ad.large",
        "name": "r5ad.large",
        "ram": 16384
    },
    "r5ad.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.xlarge",
            "instancesku": "HZJ8V6JGDWGUM5QF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5ad.xlarge",
        "name": "r5ad.xlarge",
        "ram": 32768
    },
    "r5b.12xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.12xlarge",
            "instancesku": "7ZXNNN5Z4DMKZ66A",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5b.12xlarge",
        "name": "r5b.12xlarge",
        "ram": 393216
    },
    "r5b.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.16xlarge",
            "instancesku": "6U7DBQWKEUPM26YD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5b.16xlarge",
        "name": "r5b.16xlarge",
        "ram": 524288
    },
    "r5b.24xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.24xlarge",
            "instancesku": "PDJDMA2YH5M6XBP6",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5b.24xlarge",
        "name": "r5b.24xlarge",
        "ram": 786432
    },
    "r5b.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 12.5 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5b.2xlarge",
        "name": "r5b.2xlarge",
        "ram": 65536
    },
    "r5b.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.4xlarge",
            "instancesku": "SA38FHJSUCZ3P2Z7",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5b.4xlarge",
        "name": "r5b.4xlarge",
        "ram": 131072
    },
    "r5b.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5b.8xlarge",
        "name": "r5b.8xlarge",
        "ram": 262144
    },
    "r5b.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5b.large",
        "name": "r5b.large",
        "ram": 16384
    },
    "r5b.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5b.xlarge",
        "name": "r5b.xlarge",
        "ram": 32768
    },
    "r5d.12xlarge": {
        "bandwidth": 10,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "168",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.12xlarge",
            "instancesku": "JD26ZJK92H6X64TZ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5d.12xlarge",
        "name": "r5d.12xlarge",
        "ram": 393216
    },
    "r5d.16xlarge": {
        "bandwidth": 20,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "256",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.16xlarge",
            "instancesku": "XJ5CFKTMG2X7XGD7",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5d.16xlarge",
        "name": "r5d.16xlarge",
        "ram": 524288
    },
    "r5d.24xlarge": {
        "bandwidth": 25,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5d.24xlarge",
        "name": "r5d.24xlarge",
        "ram": 786432
    },
    "r5d.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "37",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.2xlarge",
            "instancesku": "75EXHEYU7R3CEX8D",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5d.2xlarge",
        "name": "r5d.2xlarge",
        "ram": 65536
    },
    "r5d.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "70",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.4xlarge",
            "instancesku": "59NGHNMNHVTPK4SR",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5d.4xlarge",
        "name": "r5d.4xlarge",
        "ram": 131072
    },
    "r5d.8xlarge": {
        "bandwidth": 10,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "128",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.8xlarge",
            "instancesku": "QCJ53WA9MJMS6BDC",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5d.8xlarge",
        "name": "r5d.8xlarge",
        "ram": 262144
    },
    "r5d.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.large",
            "instancesku": "HE833VNWFWBDRWT4",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5d.large",
        "name": "r5d.large",
        "ram": 16384
    },
    "r5d.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "19",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.xlarge",
            "instancesku": "BAWMAEZ3ZXMUR3XH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5d.xlarge",
        "name": "r5d.xlarge",
        "ram": 32768
    },
    "r5dn.12xlarge": {
        "bandwidth": 50,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.12xlarge",
            "instancesku": "KP8X3ZRFBR5BE2QG",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 GB NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5dn.12xlarge",
        "name": "r5dn.12xlarge",
        "ram": 393216
    },
    "r5dn.16xlarge": {
        "bandwidth": 75,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.16xlarge",
            "instancesku": "9CC8VC9DH37UZAWT",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5dn.16xlarge",
        "name": "r5dn.16xlarge",
        "ram": 524288
    },
    "r5dn.24xlarge": {
        "bandwidth": 100,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.24xlarge",
            "instancesku": "8PAUA987J4Q8MTRH",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5dn.24xlarge",
        "name": "r5dn.24xlarge",
        "ram": 786432
    },
    "r5dn.2xlarge": {
        "bandwidth": 25,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5dn.2xlarge",
        "name": "r5dn.2xlarge",
        "ram": 65536
    },
    "r5dn.4xlarge": {
        "bandwidth": 25,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.4xlarge",
            "instancesku": "G3XEC6XSVJBFRRKX",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5dn.4xlarge",
        "name": "r5dn.4xlarge",
        "ram": 131072
    },
    "r5dn.8xlarge": {
        "bandwidth": 25,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5dn.8xlarge",
        "name": "r5dn.8xlarge",
        "ram": 262144
    },
    "r5dn.large": {
        "bandwidth": 25,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.large",
            "instancesku": "K4T5UGP4HDGJPFH7",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5dn.large",
        "name": "r5dn.large",
        "ram": 16384
    },
    "r5dn.xlarge": {
        "bandwidth": 25,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5dn.xlarge",
        "name": "r5dn.xlarge",
        "ram": 32768
    },
    "r5n.12xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5n.12xlarge",
        "name": "r5n.12xlarge",
        "ram": 393216
    },
    "r5n.16xlarge": {
        "bandwidth": 75,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5n.16xlarge",
        "name": "r5n.16xlarge",
        "ram": 524288
    },
    "r5n.24xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5n.24xlarge",
        "name": "r5n.24xlarge",
        "ram": 786432
    },
    "r5n.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5n.2xlarge",
        "name": "r5n.2xlarge",
        "ram": 65536
    },
    "r5n.4xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.4xlarge",
            "instancesku": "F6MSS2DTVUQR6VVK",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5n.4xlarge",
        "name": "r5n.4xlarge",
        "ram": 131072
    },
    "r5n.8xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.8xlarge",
            "instancesku": "RM86NRBPZ5CDPPRU",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5n.8xlarge",
        "name": "r5n.8xlarge",
        "ram": 262144
    },
    "r5n.large": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.large",
            "instancesku": "6WD6N47RFBKDZDF8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5n.large",
        "name": "r5n.large",
        "ram": 16384
    },
    "r5n.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "r5n.xlarge",
        "name": "r5n.xlarge",
        "ram": 32768
    },
    "r6g.12xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6g.12xlarge",
        "name": "r6g.12xlarge",
        "ram": 393216
    },
    "r6g.16xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "18000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.16xlarge",
            "instancesku": "6TECJVQ38AJEB8QM",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6g.16xlarge",
        "name": "r6g.16xlarge",
        "ram": 524288
    },
    "r6g.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.2xlarge",
            "instancesku": "G7GDESM5VCN79HV9",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6g.2xlarge",
        "name": "r6g.2xlarge",
        "ram": 65536
    },
    "r6g.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.4xlarge",
            "instancesku": "AWVHFPP2ZCKU6H6Y",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6g.4xlarge",
        "name": "r6g.4xlarge",
        "ram": 131072
    },
    "r6g.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.8xlarge",
            "instancesku": "JU3XS3CSJVG9PG6T",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6g.8xlarge",
        "name": "r6g.8xlarge",
        "ram": 262144
    },
    "r6g.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.large",
            "instancesku": "3F7S3UNTHN3MP78B",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6g.large",
        "name": "r6g.large",
        "ram": 16384
    },
    "r6g.medium": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.medium",
            "instancesku": "94RYVQHX4TWDPAUG",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6g.medium",
        "name": "r6g.medium",
        "ram": 8192
    },
    "r6g.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6g.xlarge",
        "name": "r6g.xlarge",
        "ram": 32768
    },
    "r6gd.12xlarge": {
        "bandwidth": 20,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6gd.12xlarge",
        "name": "r6gd.12xlarge",
        "ram": 393216
    },
    "r6gd.16xlarge": {
        "bandwidth": 25,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "18000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6gd.16xlarge",
        "name": "r6gd.16xlarge",
        "ram": 524288
    },
    "r6gd.2xlarge": {
        "bandwidth": 10,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.2xlarge",
            "instancesku": "NKKKMUC8VTHJGNJ3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6gd.2xlarge",
        "name": "r6gd.2xlarge",
        "ram": 65536
    },
    "r6gd.4xlarge": {
        "bandwidth": 10,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.4xlarge",
            "instancesku": "ZRRABC3BGW9ADYVP",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6gd.4xlarge",
        "name": "r6gd.4xlarge",
        "ram": 131072
    },
    "r6gd.8xlarge": {
        "bandwidth": 10,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.8xlarge",
            "instancesku": "29TFCYKSEGWR8SU4",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6gd.8xlarge",
        "name": "r6gd.8xlarge",
        "ram": 262144
    },
    "r6gd.large": {
        "bandwidth": 10,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.large",
            "instancesku": "X9MV6RERD7HKEZ65",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6gd.large",
        "name": "r6gd.large",
        "ram": 16384
    },
    "r6gd.medium": {
        "bandwidth": 10,
        "disk": 59,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.medium",
            "instancesku": "6UT3UMQBXYMNWFKU",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 59 NVMe SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6gd.medium",
        "name": "r6gd.medium",
        "ram": 8192
    },
    "r6gd.xlarge": {
        "bandwidth": 10,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "r6gd.xlarge",
        "name": "r6gd.xlarge",
        "ram": 32768
    },
    "t1.micro": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "No",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Micro instances",
            "instanceType": "t1.micro",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "0.613 GiB",
            "networkPerformance": "Very Low",
            "normalizationSizeFactor": "0.5",
            "physicalProcessor": "Variable",
            "processorArchitecture": "32-bit or 64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "t1.micro",
        "name": "t1.micro",
        "ram": 627
    },
    "t2.2xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t2.2xlarge",
            "instancesku": "WE752M8NMJH6UTAF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "t2.2xlarge",
        "name": "t2.2xlarge",
        "ram": 32768
    },
    "t2.large": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t2.large",
            "instancesku": "PAKV95ZHG77BFWBG",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Low to Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t2.large",
        "name": "t2.large",
        "ram": 8192
    },
    "t2.medium": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t2.medium",
            "instancesku": "DY99AWFYB6SX74KD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Low to Moderate",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t2.medium",
        "name": "t2.medium",
        "ram": 4096
    },
    "t2.micro": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t2.micro",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1 GiB",
            "networkPerformance": "Low to Moderate",
            "normalizationSizeFactor": "0.5",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "t2.micro",
        "name": "t2.micro",
        "ram": 1024
    },
    "t2.nano": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t2.nano",
            "instancesku": "42PSJM9P9RHT899F",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "0.5 GiB",
            "networkPerformance": "Low",
            "normalizationSizeFactor": "0.25",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "t2.nano",
        "name": "t2.nano",
        "ram": 512
    },
    "t2.small": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t2.small",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Low to Moderate",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "t2.small",
        "name": "t2.small",
        "ram": 2048
    },
    "t2.xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t2.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "t2.xlarge",
        "name": "t2.xlarge",
        "ram": 16384
    },
    "t3.2xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t3.2xlarge",
            "instancesku": "CAB4TUSN8KCWWMKW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3.2xlarge",
        "name": "t3.2xlarge",
        "ram": 32768
    },
    "t3.large": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t3.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3.large",
        "name": "t3.large",
        "ram": 8192
    },
    "t3.medium": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t3.medium",
            "instancesku": "YZ5V7FVMR7UMXETK",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3.medium",
        "name": "t3.medium",
        "ram": 4096
    },
    "t3.micro": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t3.micro",
            "instancesku": "VJNS2RY9CFHRVCMH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.5",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3.micro",
        "name": "t3.micro",
        "ram": 1024
    },
    "t3.nano": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t3.nano",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "0.5 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.25",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3.nano",
        "name": "t3.nano",
        "ram": 512
    },
    "t3.small": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t3.small",
            "instancesku": "F77PBHT47MJ9J4UZ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3.small",
        "name": "t3.small",
        "ram": 2048
    },
    "t3.xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t3.xlarge",
            "instancesku": "X6S8KH4ZJATAS9WQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3.xlarge",
        "name": "t3.xlarge",
        "ram": 16384
    },
    "t3a.2xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.2xlarge",
            "instancesku": "SZ6Q3AUG2PWZ296F",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3a.2xlarge",
        "name": "t3a.2xlarge",
        "ram": 32768
    },
    "t3a.large": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.large",
            "instancesku": "9W9WUN8874RQJPYR",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3a.large",
        "name": "t3a.large",
        "ram": 8192
    },
    "t3a.medium": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.medium",
            "instancesku": "GQT58B965R2X62MH",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3a.medium",
        "name": "t3a.medium",
        "ram": 4096
    },
    "t3a.micro": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.micro",
            "instancesku": "QVK78M9BTVPADRF8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.5",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3a.micro",
        "name": "t3a.micro",
        "ram": 1024
    },
    "t3a.nano": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.nano",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "0.5 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.25",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3a.nano",
        "name": "t3a.nano",
        "ram": 512
    },
    "t3a.small": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.small",
            "instancesku": "MCQ7R7JB2DPXEKET",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3a.small",
        "name": "t3a.small",
        "ram": 2048
    },
    "t3a.xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "Variable",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.xlarge",
            "instancesku": "YUQH22HS49VU8AJK",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "t3a.xlarge",
        "name": "t3a.xlarge",
        "ram": 16384
    },
    "t4g.2xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "t4g.2xlarge",
        "name": "t4g.2xlarge",
        "ram": 32768
    },
    "t4g.large": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.large",
            "instancesku": "8N9PPA2VHETNFJ6J",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t4g.large",
        "name": "t4g.large",
        "ram": 8192
    },
    "t4g.medium": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.medium",
            "instancesku": "5PEA4EUU3HPEJ5H3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t4g.medium",
        "name": "t4g.medium",
        "ram": 4096
    },
    "t4g.micro": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.micro",
            "instancesku": "S2M8BS97FGPUHMN7",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.5",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t4g.micro",
        "name": "t4g.micro",
        "ram": 1024
    },
    "t4g.nano": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.nano",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "0.5 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.25",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t4g.nano",
        "name": "t4g.nano",
        "ram": 512
    },
    "t4g.small": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.small",
            "instancesku": "Y33CHC2PYKBQ7U2C",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "t4g.small",
        "name": "t4g.small",
        "ram": 2048
    },
    "t4g.xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "t4g.xlarge",
        "name": "t4g.xlarge",
        "ram": 16384
    },
    "u-12tb1.112xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "u-12tb1.112xlarge",
            "instancesku": "496SWG6C9RT9DJEH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "12288 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Scalable (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "448",
            "vpcnetworkingsupport": "true"
        },
        "id": "u-12tb1.112xlarge",
        "name": "u-12tb1.112xlarge",
        "ram": 12582912
    },
    "u-6tb1.112xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "u-6tb1.112xlarge",
            "instancesku": "MTWR7MAD77AF9SWK",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "6144 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Scalable (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; Intel AVX; Intel AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "448",
            "vpcnetworkingsupport": "true"
        },
        "id": "u-6tb1.112xlarge",
        "name": "u-6tb1.112xlarge",
        "ram": 6291456
    },
    "u-6tb1.56xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "u-6tb1.56xlarge",
            "instancesku": "EJMJW2985DNSYEFD",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "6144 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Scalable (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; Intel AVX; Intel AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "224",
            "vpcnetworkingsupport": "true"
        },
        "id": "u-6tb1.56xlarge",
        "name": "u-6tb1.56xlarge",
        "ram": 6291456
    },
    "u-9tb1.112xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "u-9tb1.112xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "9216 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Scalable (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "448",
            "vpcnetworkingsupport": "true"
        },
        "id": "u-9tb1.112xlarge",
        "name": "u-9tb1.112xlarge",
        "ram": 9437184
    },
    "vt1.24xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Media Accelerator Instances",
            "instanceType": "vt1.24xlarge",
            "instancesku": "ADK6CBZ878Z8CJQT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259CL",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true"
        },
        "id": "vt1.24xlarge",
        "name": "vt1.24xlarge",
        "ram": 196608
    },
    "vt1.3xlarge": {
        "bandwidth": 3125,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2375 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Media Accelerator Instances",
            "instanceType": "vt1.3xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "24 GiB",
            "networkPerformance": "3125 Megabit",
            "normalizationSizeFactor": "24",
            "physicalProcessor": "Intel Xeon Platinum 8259CL",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "12",
            "vpcnetworkingsupport": "true"
        },
        "id": "vt1.3xlarge",
        "name": "vt1.3xlarge",
        "ram": 24576
    },
    "vt1.6xlarge": {
        "bandwidth": 6250,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "instanceFamily": "Media Accelerator Instances",
            "instanceType": "vt1.6xlarge",
            "instancesku": "PY3K83PMVBEUYB32",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "48 GiB",
            "networkPerformance": "6250 Megabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Platinum 8259CL",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "24",
            "vpcnetworkingsupport": "true"
        },
        "id": "vt1.6xlarge",
        "name": "vt1.6xlarge",
        "ram": 49152
    },
    "x1.16xlarge": {
        "bandwidth": None,
        "disk": 1920,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "174.5",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1.16xlarge",
            "instancesku": "XA7V7S39QEH3YU98",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "976 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1920 SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "x1.16xlarge",
        "name": "x1.16xlarge",
        "ram": 999424
    },
    "x1.32xlarge": {
        "bandwidth": None,
        "disk": 3840,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "349",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1.32xlarge",
            "instancesku": "XJ59DS49UK4EX6TK",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1952 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1920 SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true"
        },
        "id": "x1.32xlarge",
        "name": "x1.32xlarge",
        "ram": 1998848
    },
    "x1e.16xlarge": {
        "bandwidth": 10,
        "disk": 1920,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "179",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1952 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1920 SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "x1e.16xlarge",
        "name": "x1e.16xlarge",
        "ram": 1998848
    },
    "x1e.2xlarge": {
        "bandwidth": 10,
        "disk": 240,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1000 Mbps",
            "ecu": "23",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 240 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "x1e.2xlarge",
        "name": "x1e.2xlarge",
        "ram": 249856
    },
    "x1e.32xlarge": {
        "bandwidth": 25,
        "disk": 3840,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "340",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.32xlarge",
            "instancesku": "Q6G5G6C6EJEDN5AY",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "3904 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1920 SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true"
        },
        "id": "x1e.32xlarge",
        "name": "x1e.32xlarge",
        "ram": 3997696
    },
    "x1e.4xlarge": {
        "bandwidth": 10,
        "disk": 480,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1750 Mbps",
            "ecu": "47",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 480 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "x1e.4xlarge",
        "name": "x1e.4xlarge",
        "ram": 499712
    },
    "x1e.8xlarge": {
        "bandwidth": 10,
        "disk": 960,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "91",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.8xlarge",
            "instancesku": "5ZY9WGJKJ9DF8XVF",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "976 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 960 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "x1e.8xlarge",
        "name": "x1e.8xlarge",
        "ram": 999424
    },
    "x1e.xlarge": {
        "bandwidth": 10,
        "disk": 120,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "500 Mbps",
            "ecu": "12",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.xlarge",
            "instancesku": "BZCNQCWU5XA2AGR8",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 120 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "x1e.xlarge",
        "name": "x1e.xlarge",
        "ram": 124928
    },
    "x2gd.12xlarge": {
        "bandwidth": 20,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14250 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.12xlarge",
            "instancesku": "YHYZ8JNVTJ3DYNPQ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "x2gd.12xlarge",
        "name": "x2gd.12xlarge",
        "ram": 786432
    },
    "x2gd.16xlarge": {
        "bandwidth": 25,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.16xlarge",
            "instancesku": "N35TPJKFBWBTRUM3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true"
        },
        "id": "x2gd.16xlarge",
        "name": "x2gd.16xlarge",
        "ram": 1048576
    },
    "x2gd.2xlarge": {
        "bandwidth": 10,
        "disk": 474,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.2xlarge",
            "instancesku": "QWQAU6M6EHE8FZV3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 474 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "x2gd.2xlarge",
        "name": "x2gd.2xlarge",
        "ram": 131072
    },
    "x2gd.4xlarge": {
        "bandwidth": 10,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.4xlarge",
            "instancesku": "KTMUE4YKBZGAERNT",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true"
        },
        "id": "x2gd.4xlarge",
        "name": "x2gd.4xlarge",
        "ram": 262144
    },
    "x2gd.8xlarge": {
        "bandwidth": 12,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true"
        },
        "id": "x2gd.8xlarge",
        "name": "x2gd.8xlarge",
        "ram": 524288
    },
    "x2gd.large": {
        "bandwidth": 10,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.large",
            "instancesku": "Z6BB6KXMRVW9EJ69",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "x2gd.large",
        "name": "x2gd.large",
        "ram": 32768
    },
    "x2gd.medium": {
        "bandwidth": 10,
        "disk": 59,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.medium",
            "instancesku": "8MEF6Y3MQGMTD8RY",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 59 SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true"
        },
        "id": "x2gd.medium",
        "name": "x2gd.medium",
        "ram": 16384
    },
    "x2gd.xlarge": {
        "bandwidth": 10,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "x2gd.xlarge",
        "name": "x2gd.xlarge",
        "ram": 65536
    },
    "z1d.12xlarge": {
        "bandwidth": 25,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "235",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true"
        },
        "id": "z1d.12xlarge",
        "name": "z1d.12xlarge",
        "ram": 393216
    },
    "z1d.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2333 Mbps",
            "ecu": "45",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.2xlarge",
            "instancesku": "TJN3TJCJX99J6MMW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true"
        },
        "id": "z1d.2xlarge",
        "name": "z1d.2xlarge",
        "ram": 65536
    },
    "z1d.3xlarge": {
        "bandwidth": 10,
        "disk": 450,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "64",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.3xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "24",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 450 NVMe SSD",
            "vcpu": "12",
            "vpcnetworkingsupport": "true"
        },
        "id": "z1d.3xlarge",
        "name": "z1d.3xlarge",
        "ram": 98304
    },
    "z1d.6xlarge": {
        "bandwidth": 10,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "116",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.6xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 900 NVMe SSD",
            "vcpu": "24",
            "vpcnetworkingsupport": "true"
        },
        "id": "z1d.6xlarge",
        "name": "z1d.6xlarge",
        "ram": 196608
    },
    "z1d.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2333 Mbps",
            "ecu": "12",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.large",
            "instancesku": "TC5JJ4DDT49HWM6A",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true"
        },
        "id": "z1d.large",
        "name": "z1d.large",
        "ram": 16384
    },
    "z1d.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2333 Mbps",
            "ecu": "23",
            "enhancedNetworkingSupported": "Yes",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.xlarge",
            "instancesku": "85BM84CW7NC939FV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true"
        },
        "id": "z1d.xlarge",
        "name": "z1d.xlarge",
        "ram": 32768
    }
}